#include <Arduino.h>
#include <SPIFFS.h>
#include <WiFi.h>
#include <ESPAsyncWebServer.h>
#include <ArduinoJson.h>
#include <stdio.h>
#include "esp_log.h"
#include "MAX30105.h"
#include "spo2_algorithm.h"
#include "heartRate.h"
#include <Wire.h>

#define HTTP_PORT 80
// Define input/output pins
#define REC_PIN 4
#define ADC_PIN 7
#define RED_LED_PIN 9
#define ADC_BUFFER_SIZE 5000
#define I2C_SDA 17
#define I2C_SCL 18

static const BaseType_t app_cpu = 1;
const char *WIFI_SSID = "SDNet";
const char *WIFI_PASS = "CapstoneProject";

TaskHandle_t analogDataTask_Handle;
TaskHandle_t buttonTask_Handle;
TaskHandle_t BPM_SPO2_Handle;

// Create a buffer to hold the analog readings
float buffer[ADC_BUFFER_SIZE];

MAX30105 particleSensor;
uint32_t redTemp;
uint32_t irTemp;

uint32_t irBuffer[100]; // IR LED sensor data
uint32_t redBuffer[100]; // Red LED sensor data
int32_t bufferLength; // Data length
int32_t spo2; // SPO2 value
int8_t validSPO2; // Indicator to show if the SPO2 calculation is valid
int32_t heartRate; // Heart rate value
int8_t validHeartRate; // Indicator to show if the heart rate calculation is valid

byte ledBrightness = 50; // Options: 0=Off to 255=50mA
byte sampleAverage = 4; // Options: 1, 2, 4, 8, 16, 32
byte ledMode = 2; // Options: 1 = Red only, 2 = Red + IR, 3 = Red + IR + Green
byte sampleRate = 100; // Options: 50, 100, 200, 400, 800, 1000, 1600, 3200
int pulseWidth = 411; // Options: 69, 118, 215, 411
int adcRange = 4096; // Options: 2048, 4096, 8192, 16384

// Heartbeat code
const byte RATE_SIZE = 4; //Increase this for more averaging. 4 is good.
byte rates[RATE_SIZE]; //Array of heart rates
byte rateSpot = 0;
long lastBeat = 0; //Time at which the last beat occurred
float beatsPerMinute;
int beatAvg;
int count;

byte arrayBPM[25]; // BPM value array

AsyncWebServer server(HTTP_PORT);
AsyncWebSocket ws("/ws");

//Initialize On Board LED
struct Led {
    // state variables
    uint8_t pin;
    bool    on;

    // methods
    void update() {
        digitalWrite(pin, on ? HIGH : LOW);
    }
};
Led    onboard_led = { LED_BUILTIN, false };

//Initialize SPIFFS
void initSPIFFS() {
  if (!SPIFFS.begin()) {
    printf("Cannot mount SPIFFS volume...\n");
    while (1) {
        onboard_led.on = millis() % 200 < 50;
        onboard_led.update();
    }
  }
}

//Connect to Wifi
void initWiFi() {
  WiFi.mode(WIFI_STA);
  WiFi.begin(WIFI_SSID, WIFI_PASS);
  printf("Trying to connect [%s] ", WiFi.macAddress().c_str());
  while (WiFi.status() != WL_CONNECTED) {
      printf(".");
      delay(500);
  }
  printf(" %s\n", WiFi.localIP().toString().c_str());
}

//Initialize Web Server
String processor(const String &var) {
    return String(var == "STATE" && onboard_led.on ? "on" : "off");
}

void onRootRequest(AsyncWebServerRequest *request) {
  request->send(SPIFFS, "/index.html", "text/html", false, processor);
}

void initWebServer() {
    server.on("/", onRootRequest);
    server.serveStatic("/", SPIFFS, "/");
    server.begin();
}

//Initialize Web Socket
void notifyClients() {
    const uint8_t size = JSON_OBJECT_SIZE(1);
    StaticJsonDocument<size> json;
    json["status"] = onboard_led.on ? "on" : "off";

    char buffer[17];
    size_t len = serializeJson(json, buffer);
    ws.textAll(buffer, len);
}

void handleWebSocketMessage(void *arg, uint8_t *data, size_t len) {
    AwsFrameInfo *info = (AwsFrameInfo*)arg;
    if (info->final && info->index == 0 && info->len == len && info->opcode == WS_TEXT) {

        const uint8_t size = JSON_OBJECT_SIZE(1);
        StaticJsonDocument<size> json;
        DeserializationError err = deserializeJson(json, data);
        if (err) {
            printf("deserializeJson() failed with code ");
            printf("%s\n", err.c_str());
            return;
        }

        const char *action = json["action"];
        if (strcmp(action, "toggle") == 0) {
            //led.on = !led.on;
            notifyClients();
        }

    }
}

void onEvent(AsyncWebSocket       *server,AsyncWebSocketClient *client,AwsEventType          type,void                 *arg,uint8_t              *data,size_t                len) {

    switch (type) {
        case WS_EVT_CONNECT:
            printf("WebSocket client #%u connected from %s\n", client->id(), client->remoteIP().toString().c_str());
            break;
        case WS_EVT_DISCONNECT:
            printf("WebSocket client #%u disconnected\n", client->id());
            break;
        case WS_EVT_DATA:
            handleWebSocketMessage(arg, data, len);
            break;
        case WS_EVT_PONG:
        case WS_EVT_ERROR:
            break;
    }
}

void initWebSocket() {
    ws.onEvent(onEvent);
    server.addHandler(&ws);
}

void updateClients(void *parameters){
  while(1){
      ws.cleanupClients();
      onboard_led.on = millis() % 1000 < 50;
      onboard_led.update();
  }
}

bool writeArrayToFile(float arr[], int size) {
  // Initialize SPIFFS
  if (!SPIFFS.begin()) {
    printf("Failed to mount SPIFFS\n");
    return false;
  }

  // Serialize array as JSON
  DynamicJsonDocument doc(160000);
  JsonArray array = doc.to<JsonArray>();
  for (int i = 0; i < size; i++) {
    array.add(arr[i]);
  }
  String jsonStr;
  serializeJson(array, jsonStr);

  // Write JSON string to file
  File file = SPIFFS.open("/temp.txt", "w");
  if (!file) {
    printf("Failed to open file for writing\n");
    return false;
  }
  file.print(jsonStr);
  file.close();

  return true;
}

void writeIntToFile(int var){
  // Create a DynamicJsonDocument object
  DynamicJsonDocument doc(1024);
  doc["value"] = var;
  
  // Serialize the DynamicJsonDocument object into a JSON string
  String jsonString;
  serializeJson(doc, jsonString);
  
  // Write JSON string to file
  File file = SPIFFS.open("/heartRate.json", "w");
  if (!file) {
    Serial.println("Failed to open file for writing");
  }
  file.print(jsonString);
  file.close();
}

void writeIntToFile2(int var){
  // Create a DynamicJsonDocument object
  DynamicJsonDocument doc(1024);
  doc["value"] = var;
  
  // Serialize the DynamicJsonDocument object into a JSON string
  String jsonString;
  serializeJson(doc, jsonString);
  
  // Write JSON string to file
  File file = SPIFFS.open("/SPO2.json", "w");
  if (!file) {
    Serial.println("Failed to open file for writing");
  }
  file.print(jsonString);
  file.close();
}

void checkButtonTask(void *parameter){

  for(;;){

    if( digitalRead(REC_PIN) == LOW ){
      vTaskDelay( 100 / portTICK_PERIOD_MS );
      if( digitalRead(REC_PIN) == LOW){
        vTaskResume(analogDataTask_Handle);
        digitalWrite(RED_LED_PIN, HIGH);
        vTaskSuspend(buttonTask_Handle);
        digitalWrite(RED_LED_PIN, LOW);
        printf("Wait a little bit...\n");
        // When the adc and MAX3130 is done reading, wait 10 sec before user can record again
        vTaskDelay( 10000 / portTICK_PERIOD_MS );
        printf("Ready for new sample\n");
      }
    }
    vTaskDelay( 200 / portTICK_PERIOD_MS );
  }
}

// Define the task that collects analog data
void analogDataTask(void *parameter) {
  int currentIndex = 0;
  uint16_t avg; 
  vTaskSuspend(analogDataTask_Handle);
  while(1) {
    uint16_t analogValue = analogRead(ADC_PIN);
    float voltage = (analogValue / 4095.0) * 3.3;
    buffer[currentIndex] = voltage;
    currentIndex++; 

    if (currentIndex == ADC_BUFFER_SIZE){
      currentIndex = 0;
      printf("Buffer full\n");
      // for(i = 0; i < ADC_BUFFER_SIZE; i++){
      //   Serial.print((float)buffer[i]); 
      //   Serial.print(" - ");
      //   Serial.println(i);
      // }
      
      bool status = writeArrayToFile(buffer, ADC_BUFFER_SIZE);
      printf("ADC Task has been suspended\n");
      // Start BPM/SPO2 task
      vTaskResume(BPM_SPO2_Handle);
      vTaskSuspend(analogDataTask_Handle); 
      printf("New ADC task in progress\n"); 
      currentIndex = 0; 
    }
    // Wait for a short period of time to achieve a sampling frequency of about 44kHz
    vTaskDelay(1 / portTICK_PERIOD_MS);
  }
}

void BPM_SPO2(void *parameters)
{
  for(;;){
    vTaskSuspend(BPM_SPO2_Handle);
    printf("SPO2 Task Running\n");

    bufferLength = 100; // Buffer length of 100 stores 4 seconds of samples running at 25sps

    // Read the first 100 samples, and determine the signal range
    for (byte i = 0 ; i < bufferLength ; i++){
      while (particleSensor.available() == false){ // Do we have new data?
         particleSensor.check(); // Check the sensor for new data
      }

      redBuffer[i] = particleSensor.getRed();
      irBuffer[i] = particleSensor.getIR();
      particleSensor.nextSample(); // Move to next sample

      //printf("red=");
      //printf("%d",redBuffer[i]);
      //printf(", ir=");
      //printf("%d\n",irBuffer[i]);
    }
    // Calculate heart rate and SpO2 after first 100 samples (first 4 seconds of samples)
    maxim_heart_rate_and_oxygen_saturation(irBuffer, bufferLength, redBuffer, &spo2, &validSPO2, &heartRate, &validHeartRate);
    printf("Signal Range Dertermined (MAX30101)\n");

    // Read the second 100 samples, and determine SPO2
    for (byte i = 0 ; i < bufferLength ; i++){
      while (particleSensor.available() == false) // Do we have new data?
         particleSensor.check(); // Check the sensor for new data

      redBuffer[i] = particleSensor.getRed();
      irBuffer[i] = particleSensor.getIR();
      particleSensor.nextSample(); // Move to next sample

      //printf("red=");
      //printf("%d",redBuffer[i]);
      //printf(", ir=");
      //printf("%d\n",irBuffer[i]);
    }
    // Calculate heart rate and SpO2 after first 100 samples (first 4 seconds of samples)
    maxim_heart_rate_and_oxygen_saturation(irBuffer, bufferLength, redBuffer, &spo2, &validSPO2, &heartRate, &validHeartRate);
    
    // Read the third set of samples, and determine HR
    int i = 0;
    while(i < 25){
      while (particleSensor.available() == false) // Do we have new data?
         particleSensor.check(); // Check the sensor for new data

      long irValue = particleSensor.getIR();
      particleSensor.nextSample(); // Move to next sample

      //printf("ir=");
      //printf("%d\n",irValue);

      if (checkForBeat(irValue) == true){
        //We sensed a beat!
        long delta = millis() - lastBeat;
        lastBeat = millis();
        beatsPerMinute = 60 / (delta / 1000.0);
        arrayBPM[i] = (byte)beatsPerMinute;
        printf("%d", i);
        printf(" - ");
        printf("%d\n",arrayBPM[i]);
        i++;
      }
    }

    //Take average of readings
    beatAvg = 0;
    count = 0;
    for (int x = 0 ; x < 25 ; x++){
      if(arrayBPM[x]>44 && arrayBPM[x]<151){
        beatAvg += arrayBPM[x];
        count++;
      }  
    }
    if (beatAvg == 0) {
      beatAvg = -1;
      count = 1;  
    }
    beatAvg /= count;
    printf("Average computation done\n");

    maxim_heart_rate_and_oxygen_saturation(irBuffer, bufferLength, redBuffer, &spo2, &validSPO2, &heartRate, &validHeartRate);

    writeIntToFile(beatAvg);
    writeIntToFile2(spo2);
    printf("Integers sent over wifi");
    printf(" - bpm= %d", beatAvg);
    printf(", spo2= %d\n", spo2);
    vTaskResume(buttonTask_Handle);
  }
}

void setup() {
  // put your setup code here, to run once:
  pinMode(RED_LED_PIN, OUTPUT);
  pinMode(REC_PIN, INPUT); 
  delay(2000);
  digitalWrite(RED_LED_PIN, HIGH);
  Serial.begin(115200); 
  delay(8000);
  digitalWrite(RED_LED_PIN, LOW);
  
  Wire.begin(I2C_SDA, I2C_SCL);
  if (!particleSensor.begin(Wire, I2C_SPEED_FAST, 0x57)) // Initialize sensor
  {
    Serial.println("MAX30101 was not found.");
    while(1);
  }
  particleSensor.setup(ledBrightness, sampleAverage, ledMode, sampleRate, pulseWidth, adcRange); // Configure sensor

  initSPIFFS();
  initWiFi();
  initWebSocket();
  initWebServer();

  xTaskCreatePinnedToCore(analogDataTask, "ECG-ADC", 8000, NULL, 2, &analogDataTask_Handle, app_cpu);
  xTaskCreatePinnedToCore(updateClients, "Update Wifi", 8000, NULL, 1, NULL, 0);
  xTaskCreatePinnedToCore(checkButtonTask, "ButtonTask", 5000, NULL, 3, &buttonTask_Handle, app_cpu); 
  xTaskCreatePinnedToCore(BPM_SPO2, "BPM_SPO2_Calc", 8000, NULL, 1, &BPM_SPO2_Handle, app_cpu);
  vTaskDelete(NULL);
}


void loop() {
  // put your main code here, to run repeatedly:
}